/**
 * Copyright (c) 2014-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit;

import java.util.logging.LogRecord;

import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.TraceCode;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.logger.ProvideFlowStepFailureLogData;

/**
 * A set of assertion methods useful for writing unit test about Petals CDK.
 * 
 * @author Chrstophe DENEUX - Linagora
 * 
 */
public class Assert extends org.junit.Assert {

    /**
     * Asserts that a log record is a MONIT trace associated to the start of a service provider. If it isn't it throws
     * an {@link AssertionError}.
     * 
     * @param expectedInterface
     *            The expected interface name of the service provider
     * @param expectedService
     *            The expected service name of the service provider
     * @param expectedEndpoint
     *            The expected endpoint name of the service provider
     * @param expectedOperation
     *            The expected name of the operation invoked on the service provider
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderBeginLog(final QName expectedInterface, final QName expectedService,
            final String expectedEndpoint, final QName expectedOperation, final LogRecord logRecord) {

        return assertMonitProviderBeginLog(null, expectedInterface, expectedService, expectedEndpoint,
                expectedOperation, logRecord);
    }

    /**
     * Asserts that a log record is a MONIT trace associated to the start of a service provider. If it isn't it throws
     * an {@link AssertionError}.
     * 
     * @param expectedInterface
     *            The expected interface name of the service provider.
     * @param expectedService
     *            The expected service name of the service provider.
     * @param expectedEndpoint
     *            The expected endpoint name of the service provider. It will be ignored if it's equals to the value of
     *            {@link Constants.Component#AUTOGENERATED_ENDPOINT_NAME}.
     * @param expectedOperation
     *            The expected name of the operation invoked on the service provider.
     * @param logRecord
     *            The actual log record to check.
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderBeginLog(final FlowLogData previousFlowLogData,
            final QName expectedInterface, final QName expectedService,
            final String expectedEndpoint, final QName expectedOperation, final LogRecord logRecord) {

        final FlowLogData flowLogData = Assert.checkCommonsAttribute(TraceCode.PROVIDE_FLOW_STEP_BEGIN,
                previousFlowLogData != null
                        ? (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME) : null,
                null, previousFlowLogData != null
                        ? (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME) : null,
                logRecord);

        assertEquals("Unexpected interface name", expectedInterface == null ? null : expectedInterface.toString(),
                flowLogData.get(FlowLogData.FLOW_STEP_INTERFACE_NAME_PROPERTY_NAME));
        assertEquals("Unexpected service name", expectedService == null ? null : expectedService.toString(),
                flowLogData.get(FlowLogData.FLOW_STEP_SERVICE_NAME_PROPERTY_NAME));
        // no need to check it if's autogenerated...
        if (!Constants.Component.AUTOGENERATED_ENDPOINT_NAME.equals(expectedEndpoint)) {
            assertEquals("Unexpected endpoint name", expectedEndpoint,
                    flowLogData.get(FlowLogData.FLOW_STEP_ENDPOINT_NAME_PROPERTY_NAME));
        }
        assertEquals("Unexpected operation name", expectedOperation == null ? null : expectedOperation.toString(),
                flowLogData.get(FlowLogData.FLOW_STEP_OPERATION_NAME_PROPERTY_NAME));

        return flowLogData;

    }

    /**
     * <p>
     * Asserts that a log record is a MONIT trace associated to the start of an external service provider. If it isn't
     * it throws an {@link AssertionError}.
     * </p>
     * <p>
     * Checks:
     * </p>
     * <ul>
     * <li>the trace code of the log record is {@link TraceCode#PROVIDE_EXT_FLOW_STEP_BEGIN},</li>
     * <li>the flow instance identifier is one of the previous step,</li>
     * <li>the flow step identifier is a new one,</li>
     * <li>the previous flow step identifier is the flow step identifier of the previous step,</li>
     * <li>a message exchange is included (for potential dump through the right log handler)</li>
     * </ul>
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the previous step
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderExtBeginLog(final FlowLogData previousFlowLogData,
            final LogRecord logRecord) {

        return Assert.checkCommonsAttribute(TraceCode.PROVIDE_EXT_FLOW_STEP_BEGIN,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME), null,
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), logRecord);
    }

    /**
     * <p>
     * Asserts that a log record is a MONIT trace associated to the end of an external service provider. If it isn't it
     * throws an {@link AssertionError}.
     * </p>
     * <p>
     * Checks:
     * </p>
     * <ul>
     * <li>the trace code of the log record is {@link TraceCode#PROVIDE_EXT_FLOW_STEP_END},</li>
     * <li>the flow instance identifier is one of the previous step,</li>
     * <li>the flow step identifier is one of the previous step,</li>
     * <li>the previous flow step identifier has no sens,</li>
     * <li>a message exchange is included (for potential dump through the right log handler)</li>
     * </ul>
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the step startup
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderExtEndLog(final FlowLogData previousFlowLogData,
            final LogRecord logRecord) {

        return Assert.checkCommonsAttribute(TraceCode.PROVIDE_EXT_FLOW_STEP_END,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);
    }

    /**
     * Asserts that a log record is a MONIT trace associated to the failure of an external service provider. If it isn't
     * it throws an {@link AssertionError}.
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the step startup
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderExtFailureLog(final FlowLogData previousFlowLogData,
            final LogRecord logRecord) {

        final FlowLogData flowLogData = Assert.checkCommonsAttribute(TraceCode.PROVIDE_EXT_FLOW_STEP_FAILURE,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);

        assertNotNull("Failure message is missing",
                flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME));
        assertFalse("Empty failure message",
                ((String) flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME)).isEmpty());
        return flowLogData;
    }

    /**
     * <p>
     * Asserts that a log record is a MONIT trace associated to the end of a service provider. If it isn't it throws an
     * {@link AssertionError}.
     * </p>
     * <p>
     * Checks:
     * </p>
     * <ul>
     * <li>the trace code of the log record is {@link TraceCode#PROVIDE_FLOW_STEP_END},</li>
     * <li>the flow instance identifier is one of the previous step,</li>
     * <li>the flow step identifier is one of the previous step,</li>
     * <li>the previous flow step identifier has no sens,</li>
     * <li>a message exchange is included (for potential dump through the right log handler)</li>
     * </ul>
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the step startup
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderEndLog(final FlowLogData previousFlowLogData, final LogRecord logRecord) {

        return Assert.checkCommonsAttribute(TraceCode.PROVIDE_FLOW_STEP_END,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);
    }

    /**
     * Asserts that a log record is a MONIT trace associated to the failure of a service provider. If it isn't it throws
     * an {@link AssertionError}.
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the step startup
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitProviderFailureLog(final FlowLogData previousFlowLogData, final LogRecord logRecord) {

        final FlowLogData flowLogData = Assert.checkCommonsAttribute(TraceCode.PROVIDE_FLOW_STEP_FAILURE,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);

        assertNotNull("Failure message is missing",
                flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME));
        assertFalse("Empty failure message",
                ((String) flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME)).isEmpty());
        
        return flowLogData;
    }

    /**
     * Asserts that a log record is a MONIT trace associated to the start of a service consumer. If it isn't it throws
     * an {@link AssertionError}.
     * <p>
     * Checks:
     * </p>
     * <ul>
     * <li>the trace code of the log record is {@link TraceCode#CONSUME_EXT_FLOW_STEP_BEGIN},</li>
     * <li>the flow instance identifier is defined,</li>
     * <li>the flow step identifier is defined,</li>
     * <li>the previous flow step identifier has no sens,</li>
     * <li>a message exchange is included (for potential dump through the right log handler).</li>
     * </ul>
     * 
     * @param logRecord
     *            The actual log record to check.
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitConsumerExtBeginLog(final LogRecord logRecord) {

        return Assert.checkCommonsAttribute(TraceCode.CONSUME_EXT_FLOW_STEP_BEGIN, null, null, null, logRecord);
    }

    /**
     * Asserts that a log record is a MONIT trace associated to the end of a service consumer. If it isn't it throws an
     * {@link AssertionError}.
     * 
     * @param previousFlowLogData
     *            The flow attributes of the MONIT trace associated to the step startup
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData assertMonitConsumerExtEndLog(final FlowLogData previousFlowLogData,
            final LogRecord logRecord) {

        return Assert.checkCommonsAttribute(TraceCode.CONSUME_EXT_FLOW_STEP_END,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);
    }

    public static FlowLogData assertMonitConsumerExtFailureLog(final FlowLogData previousFlowLogData,
            final LogRecord logRecord) {

        final FlowLogData flowLogData = Assert.checkCommonsAttribute(TraceCode.CONSUME_EXT_FLOW_STEP_FAILURE,
                (String) previousFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                (String) previousFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME), null, logRecord);

        assertNotNull("Failure message is missing",
                flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME));
        assertFalse("Empty failure message",
                ((String) flowLogData.get(ProvideFlowStepFailureLogData.FLOW_STEP_FAILURE_MESSAGE_NAME)).isEmpty());

        return flowLogData;
    }

    /**
     * <p>
     * Common checks about all MONIT traces:
     * </p>
     * <ul>
     * <li>the trace code of the log record is the given one,</li>
     * <li>the flow instance identifier is the given one, or is not <code>null</code>,</li>
     * <li>the flow step identifier is the given one, or is not <code>null</code>,</li>
     * <li>the previous flow step identifier is the given one,</li>
     * <li>a message exchange is included (for potential dump through the right log handler)</li>
     * </ul>
     * 
     * @param expectedTraceCode
     *            The expected trace code of the MONIT trace
     * @param expectedFlowInstanceId
     *            The expected flow instance identifier of the MONIT trace, or <code>null</code> if the expected flow
     *            instance identifier is not known and must be set
     * @param expectedFlowStepId
     *            The expected flow step identifier of the MONIT trace, or <code>null</code> if the expected flow step
     *            identifier is not known and must be set
     * @param expectedFlowPreviousStepId
     *            The expected flow previous step identifier of the MONIT trace or <code>null</code> if the previous
     *            flow step identifier has no sens
     * @param logRecord
     *            The actual log record to check
     * @return The flow log data of the log record
     */
    public static FlowLogData checkCommonsAttribute(final TraceCode expectedTraceCode,
            final String expectedFlowInstanceId, final String expectedFlowStepId,
            final String expectedFlowPreviousStepId, final LogRecord logRecord) {

        assertEquals(Level.MONIT, logRecord.getLevel());

        final Object[] parameters = logRecord.getParameters();
        assertNotNull("The log record does not contain parameters!", parameters);
        assertEquals("Unexpected number of parameters into the log record!", 1, parameters.length);
        final Object traceDetails = parameters[0];
        assertNotNull("The parameter of the log record does not contain trace details!", traceDetails);
        assertTrue("Unexpected trace detail object!", traceDetails instanceof FlowLogData);

        final FlowLogData flowLogData = (FlowLogData) traceDetails;

        assertEquals("Unexpected trace code", expectedTraceCode, flowLogData.get(FlowLogData.FLOW_STEP_TRACE_CODE));
        if (expectedFlowInstanceId == null) {
            assertNotNull("Flow instance identifier is missing",
                    flowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME));
        } else {
            assertEquals("Unexpected flow instance identifier", expectedFlowInstanceId,
                    flowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME));
        }
        if (expectedFlowStepId == null) {
            assertNotNull("Flow step identifier is missing",
                    flowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME));
            if (expectedFlowPreviousStepId != null) {
                assertNotEquals(flowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME),
                        expectedFlowPreviousStepId);
            }
        } else {
            assertEquals("Unexpected flow step identifier", expectedFlowStepId,
                    flowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME));
        }
        if (expectedFlowPreviousStepId != null) {
            assertEquals("Unexpected flow previous step identifier", expectedFlowPreviousStepId,
                    flowLogData.get(FlowLogData.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME));
        }

        // TODO: Remove the following hack about correlated flow attributes when a right model of flow steps will be
        // designed and taken into account the BPMN process use case where no exchange is associated to traces about
        // process instance creation, user task completion and call activity steps.
        if (flowLogData.get("correlatedFlowInstanceId") == null && flowLogData.get("callActivityInstanceId") == null) {
            if (expectedTraceCode == TraceCode.PROVIDE_FLOW_STEP_BEGIN
                    || expectedTraceCode == TraceCode.PROVIDE_FLOW_STEP_END
                    || expectedTraceCode == TraceCode.PROVIDE_FLOW_STEP_FAILURE) {
                assertNotNull("Message exchange instance is missing",
                        flowLogData.get(FlowLogData.FLOW_EXCHANGE_PROPERTY_NAME));
            }
        }

        return flowLogData;
    }

    /**
     * Asserts that the flow instance identifiers are equal. If they are not, an {@link AssertionError} is thrown. If
     * flow instance identifiers of <code>expectedFlowLogData</code> and <code>actualFlowAttributes</code> are
     * <code>null</code>, they are considered equal.
     *
     * @param expectedFlowLogData
     *            The flow log data containing the expected flow instance identifier
     * @param actualFlowAttributes
     *            The flow attributes containing the flow instance identifier to check against the one of
     *            <code>expectedFlowLogData</code>
     */
    public static void assertMonitFlowInstanceIdEquals(final FlowLogData expectedFlowLogData,
            final FlowAttributes actualFlowAttributes) {
        assertEquals(expectedFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                actualFlowAttributes.getFlowInstanceId());
    }

    /**
     * Asserts that the flow instance identifiers are <b>not</b> equal. If they are, an {@link AssertionError} is
     * thrown. If flow instance identifiers of <code>expectedFlowLogData</code> and <code>actualFlowAttributes</code>
     * are <code>null</code>, they are considered equal.
     *
     * @param expectedFlowLogData
     *            The flow log data containing the expected flow instance identifier
     * @param actualFlowLogData
     *            The actual flow log data containing flow attributes to check against the ones of
     *            <code>expectedFlowLogData</code>
     */
    public static void assertMonitFlowInstanceIdNotEquals(final FlowLogData expectedFlowLogData,
            final FlowLogData actualFlowLogData) {
        assertNotEquals(expectedFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME),
                actualFlowLogData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME));
    }

    /**
     * Asserts that the flow step identifiers are equal. If they are not, an {@link AssertionError} is thrown. If flow
     * step identifiers of <code>expectedFlowLogData</code> and <code>actualFlowAttributes</code> are <code>null</code>,
     * they are considered equal.
     *
     * @param expectedFlowLogData
     *            The flow log data containing the expected flow step identifier
     * @param actualFlowAttributes
     *            The flow attributes containing the flow step identifier to check against the one of
     *            <code>expectedFlowLogData</code>
     */
    public static void assertMonitFlowStepIdEquals(final FlowLogData expectedFlowLogData,
            final FlowAttributes actualFlowAttributes) {
        assertEquals(expectedFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME),
                actualFlowAttributes.getFlowStepId());
    }

    /**
     * Asserts that the flow step identifiers are <b>not</b> equal. If they are not, an {@link AssertionError} is
     * thrown. If flow step identifiers of <code>expectedFlowLogData</code> and <code>actualFlowAttributes</code> are
     * <code>null</code>, they are considered equal.
     *
     * @param expectedFlowLogData
     *            The flow log data containing the expected flow step identifier
     * @param actualFlowAttributes
     *            The flow attributes containing the flow step identifier to check against the one of
     *            <code>expectedFlowLogData</code>
     */
    public static void assertMonitFlowStepIdNotEquals(final FlowLogData expectedFlowLogData,
            final FlowAttributes actualFlowAttributes) {
        assertNotEquals(expectedFlowLogData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME),
                actualFlowAttributes.getFlowStepId());
    }

}
