/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl.message;

import java.util.Collections;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public abstract class AbstractWrappedMessage implements Message {

    protected PetalsMessageExchange messageExchange;
    
    public AbstractWrappedMessage(final PetalsMessageExchange messageExchange) {
        this.messageExchange = messageExchange;
    }
    
    @Override
    public String getFlowStepInterfaceName() {
        return String.valueOf(this.messageExchange.getInterfaceName());
    }

    @Override
    public String getFlowStepServiceName() {
        return String.valueOf(this.messageExchange.getEndpoint().getServiceName());
    }

    @Override
    public String getFlowStepOperationName() {
        return String.valueOf(this.messageExchange.getOperation());
    }

    @Override
    public String getFlowStepEndpointName() {
        return this.messageExchange.getEndpoint().getEndpointName();
    }

    @Override
    public String getFlowStepId() {
        return FlowAttributesExchangeHelper.getFlowAttributes(messageExchange).getFlowStepId();
    }
    
    @Override
    public MessageExchange getMessageExchange() {
        return this.messageExchange;
    }
    
    @Override
    public boolean isInOnly() {
        return MEPPatternConstants.IN_ONLY.equals(this.messageExchange.getPattern());
    }

    @Override
    public boolean isInOptionalOut() {
        return MEPPatternConstants.IN_OPTIONAL_OUT.equals(this.messageExchange.getPattern());
    }

    @Override
    public boolean isInOut() {
        return MEPPatternConstants.IN_OUT.equals(this.messageExchange.getPattern());
    }

    @Override
    public boolean isRobustInOnly() {
        return MEPPatternConstants.ROBUST_IN_ONLY.equals(this.messageExchange.getPattern());
    }

    protected final Source getPayload(final String messageName) {
        if (PetalsMessageExchange.FAULT_MSG.equals(messageName)) {
            final Fault fault = this.messageExchange.getFault();
            if (fault != null) {
                return fault.getContent();
            } else {
                return null;
            }
        } else {
            final NormalizedMessage message = this.messageExchange.getMessage(messageName);
            if (message != null) {
                return message.getContent();
            } else {
                return null;
            }
        }
	}

    /**
     * Retrieve an attachment on the given message
     * 
     * @param messageName
     *            The message on which the attachment must be retrieved
     * @param attachmentName
     *            The name of the attachment to retrieve
     * @return The attachment matching the given name, or {@code null} is no attachment matching the given name exists.
     */
    protected final DataHandler getAttachment(final String messageName, final String attachmentName) {
        if (PetalsMessageExchange.FAULT_MSG.equals(messageName)) {
            final Fault fault = this.messageExchange.getFault();
            if (fault != null) {
                return fault.getAttachment(attachmentName);
            } else {
                return null;
            }
        } else {
            final NormalizedMessage message = this.messageExchange.getMessage(messageName);
            if (message != null) {
                return message.getAttachment(attachmentName);
            } else {
                return null;
            }
        }
    }

    /**
     * @param messageName
     *            The message on which the attachment names must be retrieved
     * @return The attachment names available on the given message, or an empty collection if no attachment is available
     *         for the given message, or {@code null} if the given message does not exist.
     */
    protected final Set<String> getAttachmentNames(final String messageName) {
        if (PetalsMessageExchange.FAULT_MSG.equals(messageName)) {
            final Fault fault = this.messageExchange.getFault();
            if (fault != null) {
                final Set<String> attachments = fault.getAttachmentNames();
                return (Set<String>) (attachments == null ? Collections.emptySet() : attachments);
            } else {
                return null;
            }
        } else {
            final NormalizedMessage message = this.messageExchange.getMessage(messageName);
            if (message != null) {
                final Set<String> attachments = message.getAttachmentNames();
                return (Set<String>) (attachments == null ? Collections.emptySet() : attachments);
            } else {
                return null;
            }
        }
    }

    /**
     * 
     * @param messageName
     *            The message on which the attachment must be added
     * @param attachmentName
     *            The name of the attachment to add
     * @param attachment
     *            The attachment to add
     * @return {@code true} if the attachment has been added to the given message
     * @throws MessagingException
     */
    protected final boolean addAttachment(final String messageName, final String attachmentName,
            final DataHandler attachment) throws MessagingException {
        if (PetalsMessageExchange.FAULT_MSG.equals(messageName)) {
            final Fault fault = this.messageExchange.getFault();
            if (fault != null) {
                fault.addAttachment(attachmentName, attachment);
                return true;
            } else {
                return false;
            }
        } else {
            final NormalizedMessage message = this.messageExchange.getMessage(messageName);
            if (message != null) {
                message.addAttachment(attachmentName, attachment);
                return true;
            } else {
                return false;
            }
        }
    }

    @Override
    public Source getFault() {
        return this.getPayload(PetalsMessageExchange.FAULT_MSG);
    }

    @Override
    public Source getIn() {
        return this.getPayload(PetalsMessageExchange.IN_MSG);
    }

    @Override
    public Source getOut() {
        return this.getPayload(PetalsMessageExchange.OUT_MSG);
    }

    @Override
    public DataHandler getFaultAttachment(final String attachmentName) {
        return this.getAttachment(PetalsMessageExchange.FAULT_MSG, attachmentName);
    }

    @Override
    public DataHandler getInAttachment(final String attachmentName) {
        return this.getAttachment(PetalsMessageExchange.IN_MSG, attachmentName);
    }

    @Override
    public DataHandler getOutAttachment(final String attachmentName) {
        return this.getAttachment(PetalsMessageExchange.OUT_MSG, attachmentName);
    }

    @Override
    public boolean addFaultAttachment(final String attachmentName, final DataHandler attachment)
            throws MessagingException {
        return this.addAttachment(PetalsMessageExchange.FAULT_MSG, attachmentName, attachment);
    }

    @Override
    public boolean addInAttachment(final String attachmentName, final DataHandler attachment)
            throws MessagingException {
        return this.addAttachment(PetalsMessageExchange.IN_MSG, attachmentName, attachment);
    }

    @Override
    public boolean addOutAttachment(final String attachmentName, final DataHandler attachment)
            throws MessagingException {
        return this.addAttachment(PetalsMessageExchange.OUT_MSG, attachmentName, attachment);
    }

    @Override
    public Set<String> getFaultAttachmentNames() {
        return this.getAttachmentNames(PetalsMessageExchange.FAULT_MSG);
    }

    @Override
    public Set<String> getInAttachmentNames() {
        return this.getAttachmentNames(PetalsMessageExchange.IN_MSG);
    }

    @Override
    public Set<String> getOutAttachmentNames() {
        return this.getAttachmentNames(PetalsMessageExchange.OUT_MSG);
    }

    @Override
    public Exception getError() {
        return this.messageExchange.getError();
    }

    @Override
    public ExchangeStatus getStatus() {
        return this.messageExchange.getStatus();
    }
}
