/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl.message;

import java.io.InputStream;
import java.io.StringReader;
import java.net.URI;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.junit.Assert;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.junit.Component;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.impl.ConsumesServiceConfiguration;
import org.ow2.petals.component.framework.junit.impl.PCServiceConfiguration;
import org.ow2.petals.component.framework.junit.impl.ServiceConfiguration;
import org.ow2.petals.component.framework.junit.impl.mock.MockServiceEndpoint;
import org.ow2.petals.component.framework.util.jaxb.JBIAttachmentMarshaller;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.xml.BytesSource;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.stream.EasyByteArrayOutputStream;

/**
 * A message exchange as request to send to a service provider
 * 
 * @author Marc Jambert - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 */
public class RequestToProviderMessage extends AbstractWrappedToComponentMessage implements RequestMessage {
    
    public RequestToProviderMessage(final Component cut, final String service, final byte[] inContent) {
        this(cut, service, null, null, new BytesSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep,
            final Object objContent, final Marshaller marshaller) {
        this(cut, service, operation, mep);

        setInContent(objContent, marshaller);

    }

    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep,
            final byte[] inContent) {
        this(cut, service, operation, mep, new BytesSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String endpointName, final QName serviceQName,
            final QName interfaceQName, final QName operation, final URI mep, final byte[] inContent) {
        this(cut, endpointName, serviceQName, interfaceQName, operation, mep, new BytesSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep,
            final InputStream inContent) {
        this(cut, service, operation, mep, inContent == null ? null : new StreamSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String endpointName, final QName serviceQName,
            final QName interfaceQName, final QName operation, final URI mep, final InputStream inContent) {
        this(cut, endpointName, serviceQName, interfaceQName, operation, mep,
                inContent == null ? null : new StreamSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String service, final InputStream inContent) {
        this(cut, service, null, null, inContent == null ? null : new StreamSource(inContent));
    }

    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep,
            final String inContent) {
        this(cut, service, operation, mep, inContent == null ? null : new StreamSource(new StringReader(inContent)));
    }

    public RequestToProviderMessage(final Component cut, final String endpointName, final QName serviceQName,
            final QName interfaceQName, final QName operation, final URI mep, final String inContent) {
        this(cut, endpointName, serviceQName, interfaceQName, operation, mep,
                inContent == null ? null : new StreamSource(new StringReader(inContent)));
    }

    public RequestToProviderMessage(final Component cut, final String service, final String inContent) {
        this(cut, service, null, null, inContent == null ? null : new StreamSource(new StringReader(inContent)));
    }

    public RequestToProviderMessage(final Component cut, final String service) {
        this(cut, service, null, null, (Source) null);
    }

    public RequestToProviderMessage(final Component cut, final String service, final Source inContent) {
        this(cut, service, null, null, inContent);
    }

    /**
     * Create a new request with an in-payload.
     * 
     * @param cut
     *            The component under test from which the service provider configuration and the message exchange
     *            factory will be retrieved.
     * @param service
     *            The name of the service provider configuration previously
     * @param operation
     *            The operation of the request
     * @param mep
     *            The pattern of the request
     * @param inContent
     *            The in-payload
     */
    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep,
            final Source inContent) {
        this(cut, service, operation, mep);

        setInContent(inContent);
    }

    public RequestToProviderMessage(final Component cut, final String endpointName, final QName serviceQName,
            final QName interfaceQName, final QName operation, final URI mep, final Source inContent) {
        this(cut, endpointName, serviceQName, interfaceQName, operation, mep);

        setInContent(inContent);
    }

    /**
     * Create a new request without payload.
     * 
     * @param cut
     *            The component under test from which the service provider configuration and the message exchange
     *            factory will be retrieved.
     * @param service
     *            The name of the service provider configuration previously
     * @param operation
     *            The operation of the request
     * @param mep
     *            The pattern of the request
     */
    public RequestToProviderMessage(final Component cut, final String service, final QName operation, final URI mep) {
        super(createMessageExchange(cut, service, operation, mep));
    }

    public RequestToProviderMessage(final Component cut, final String endpointName, final QName serviceQName,
            final QName interfaceQName, final QName operation, final URI mep) {
        super(createMessageExchange(cut, operation, mep, endpointName, serviceQName, interfaceQName));
    }

    /**
     * Create a new request from the given one that will be sent to the given service provider. The following
     * information will be retrieved from the given request: operation, pattern, in-payload, attachments
     * 
     * @param cut
     *            The component under test from which the service provider configuration and the message exchange
     *            factory will be retrieved.
     * @param service
     *            The name of the service provider configuration previously
     * @param request
     *            The request from which the operation, pattern, in-payload and attachments will be copied
     * @throws MessagingException
     */
    public RequestToProviderMessage(final Component cut, final String service, final RequestMessage request)
            throws MessagingException {
        this(cut, service, request.getMessageExchange().getOperation(), request.getMessageExchange().getPattern(),
                request.getIn());
        for (final String attachmentName : request.getInAttachmentNames()) {
            this.addInAttachment(attachmentName, request.getInAttachment(attachmentName));
        }
    }

    private void setInContent(final Object objContent, final Marshaller marshaller) {

        final EasyByteArrayOutputStream ebaos = new EasyByteArrayOutputStream();
        try {
            final NormalizedMessage inNm = this.messageExchange.createMessage();
            marshaller.setProperty(Marshaller.JAXB_ENCODING, "UTF-8");
            marshaller.setAttachmentMarshaller(new JBIAttachmentMarshaller(inNm));
            marshaller.marshal(objContent, ebaos);
            inNm.setContent(new BytesSource(ebaos.toRawByteArray()));
            this.messageExchange.setMessage(inNm, Exchange.IN_MESSAGE_NAME);
        } catch (final MessagingException | JAXBException e) {
            throw new UncheckedException("Impossible to create a message exchange", e);
        } finally {
            ebaos.close();
        }
    }

    private void setInContent(final Source inContent) {
        try {
            final NormalizedMessage inNm = this.messageExchange.createMessage();
            inNm.setContent(inContent);
            this.messageExchange.setMessage(inNm, Exchange.IN_MESSAGE_NAME);
        } catch (final MessagingException e) {
            throw new UncheckedException("Impossible to create a message exchange", e);
        }
    }

	@Override
    public Source getPayload() {
        return getIn();
    }

    private static PetalsMessageExchange createMessageExchange(final Component cut, final String service,
            final QName operation, final URI mep) {

        final ServiceConfiguration serviceConfiguration = cut.getServiceConfiguration(service);

        Assert.assertTrue(
                "Must use a ConsumesServiceConfiguration or a ProvidesServiceConfiguration to create an exchange",
                serviceConfiguration instanceof PCServiceConfiguration);

        final QName realOperation;
        if (operation == null) {
            Assert.assertTrue(
                    "If creating a Message without a ConsumesServiceConfiguration, please pass the operation to the constructor",
                    serviceConfiguration instanceof ConsumesServiceConfiguration);
            realOperation = ((ConsumesServiceConfiguration) serviceConfiguration).getOperation();
            Assert.assertNotNull(
                    "If creating a Message with a ServiceConfiguration without an operation, please pass it to the constructor",
                    realOperation);
        } else {
            realOperation = operation;
        }

        final URI realMep;
        if (mep == null) {
            Assert.assertTrue(
                    "If creating a Message without a ConsumesServiceConfiguration, please pass the MEP to the constructor",
                    serviceConfiguration instanceof ConsumesServiceConfiguration);
            final MEPType mept = ((ConsumesServiceConfiguration) serviceConfiguration).getMEP();
            Assert.assertNotNull(
                    "If creating a Message with a ServiceConfiguration without a MEP, please pass it to the constructor",
                    mept);
            realMep = toURI(mept);
        } else {
            realMep = mep;
        }

        final PCServiceConfiguration sc = (PCServiceConfiguration) serviceConfiguration;

        return createMessageExchange(cut, realOperation, realMep, sc.getEndpointName(), sc.getServiceName(),
                sc.getInterfaceName());
    }

    private static PetalsMessageExchange createMessageExchange(final Component cut, final QName operation,
            final URI mep, final String endpointName, final QName serviceName, final QName interfaceName) {

        final PetalsMessageExchange messageExchange;
        try {
            messageExchange = cut.getTestMessageExchangeFactory().createExchange(mep);
        } catch (final MessagingException e) {
            throw new UncheckedException(e);
        }

        if (endpointName != null && serviceName != null) {
            // the test should send to an existing endpoint of the component (because we are currently testing the
            // component, not something else)
            final MockServiceEndpoint endpoint = cut.getEndpointDirectory().resolveEndpoint(serviceName, endpointName);
            Assert.assertNotNull("No endpoint exists for " + serviceName + " and " + endpointName, endpoint);
            messageExchange.setEndpoint(endpoint);
        }

        messageExchange.setInterfaceName(interfaceName);
        messageExchange.setService(serviceName);
        messageExchange.setOperation(operation);
        return messageExchange;
    }

    private static URI toURI(final MEPType mep) {
        switch (mep) {
            case IN_ONLY:
                return MEPPatternConstants.IN_ONLY.value();
            case IN_OPTIONAL_OUT:
                return MEPPatternConstants.IN_OPTIONAL_OUT.value();
            case IN_OUT:
                return MEPPatternConstants.IN_OUT.value();
            case ROBUST_IN_ONLY:
                return MEPPatternConstants.ROBUST_IN_ONLY.value();
            default:
                throw new UncheckedException("Unkonwn MEP " + mep);
        }
    }
    
}
