/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl.mock;

import java.util.Collection;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.jbi.messaging.PetalsDeliveryChannel;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.messaging.exchange.impl.DefaultMessageExchangeFactory;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.jbi.servicedesc.endpoint.util.ServiceEndpointHelper;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * 
 * One of the responsibility of this class is to present to the tested component {@link MessageExchange} that are
 * {@link MessageExchangeWrapper}.
 * 
 * In particular it means that {@link #accept()} and {@link #accept(long)} returns {@link MessageExchangeWrapper} to the
 * tested component and {@link #send(MessageExchange)}, {@link #sendSync(MessageExchange)} and
 * {@link #sendSync(MessageExchange, long)} take {@link MessageExchangeWrapper} as parameters.
 * 
 * @author Adrien Ruffie - EBM WebSourcing
 * @author Victor Noel
 */
public class MockDeliveryChannel implements PetalsDeliveryChannel {

    private final LoggingUtil logger;

    /**
     * For messages from the testing client or other (non-testing client) services to the tested component
     */
    private final BlockingQueue<PetalsMessageExchange> incomingExchangesQueue;

    /**
     * For messages from the tested component to the testing client
     */
    private final BlockingQueue<PetalsMessageExchange> exchangesToClientQueue;

    /**
     * For messages from the tested component to other (non-testing client) services
     */
    private final BlockingQueue<PetalsMessageExchange> exchangesToServerQueue;

    private volatile boolean opened = true;

    private final Location location;

    private final MockComponentContext context;

    private final MockEndpointDirectory directory;

    public MockDeliveryChannel(final Logger logger, final String componentName, final MockEndpointDirectory directory,
            final MockComponentContext context) {
        this.directory = directory;
        this.context = context;
        this.location = new Location("MockContainer", componentName);
        this.logger = new LoggingUtil(logger);
        this.incomingExchangesQueue = new LinkedBlockingQueue<PetalsMessageExchange>();
        this.exchangesToClientQueue = new LinkedBlockingQueue<PetalsMessageExchange>();
        this.exchangesToServerQueue = new LinkedBlockingQueue<PetalsMessageExchange>();
    }

    @Override
    public ComponentContext getContext() {
        return this.context;
    }

    @Override
    public boolean isOpened() {
        return this.opened;
    }

    @Override
    public Location getLocation() {
        return this.location;
    }

    @Override
    public Logger getLogger() {
        return this.logger.getLogger();
    }

    public void enqueueIncomingExchange(final PetalsMessageExchange mex) {
        this.logger.call();
        try {

            if (mex.isNewExchange()) {
                setElectedServiceEndpoint(mex);
            } else {
                assert mex.getElectedProviderEndpoint() != null;
            }

            synchronized (mex) {
                if (mex.isProvider()) {
                    mex.setRole(Role.CONSUMER);
                } else {
                    mex.setRole(Role.PROVIDER);
                }
                if (mex.getProperty(PROPERTY_SENDSYNC) != null) {
                    mex.setProperty(PROPERTY_SENDSYNC, null);
                    mex.notifyAll();
                } else {
                    this.incomingExchangesQueue.put(mex);
                }
            }
        } catch (final InterruptedException | MessagingException e) {
            throw new UncheckedException(e);
        }
    }

    public void clearIncomingExchanges() {
        this.logger.call();
        this.incomingExchangesQueue.clear();
    }

    /**
     * @param timeout
     *            Timeout in millisecond to dequeue the message exchange. If <code>0</code>, no timeout will occur.
     * @return
     */
    public PetalsMessageExchange dequeueExchangeToServer(final long timeout) throws TimeoutException {
        this.logger.start();
        try {
            if (timeout == 0) {
                return this.exchangesToServerQueue.take();
            } else {
                PetalsMessageExchange polled = this.exchangesToServerQueue.poll(timeout, TimeUnit.MILLISECONDS);
                if (polled == null) {
                    throw new TimeoutException();
                }
                return polled;
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        } finally {
            this.logger.end();
        }
    }

    public void clearExchangesToServer() {
        this.logger.call();
        this.exchangesToServerQueue.clear();
    }

    public int getExchangesToServerCount() {
        return this.exchangesToServerQueue.size();
    }

    /**
     * @param timeout
     *            Timeout in millisecond to dequeue the message exchange. If <code>0</code>, no timeout will occur.
     * @return
     */
    public PetalsMessageExchange dequeueExchangeToClient(final long timeout) throws TimeoutException {
        this.logger.start();
        try {
            if (timeout == 0) {
                return this.exchangesToClientQueue.take();
            } else {
                PetalsMessageExchange polled = this.exchangesToClientQueue.poll(timeout, TimeUnit.MILLISECONDS);
                if (polled == null) {
                    throw new TimeoutException();
                }
                return polled;
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        } finally {
            this.logger.end();
        }

    }

    public void clearExchangesToClient() {
        this.logger.call();
        this.exchangesToClientQueue.clear();
    }

    public int getExchangesToClientCount() {
        return this.exchangesToClientQueue.size();
    }

    public int getExchangesToComponentCount() {
        return this.incomingExchangesQueue.size();
    }

    @Override
    public MessageExchangeWrapper accept() throws MessagingException {
        this.logger.start();
        try {
            return this.accept(0);
        } finally {
            this.logger.end();
        }
    }

    @Override
    public MessageExchangeWrapper accept(final long timeoutMS) throws MessagingException {
        this.logger.start();
        try {
            final PetalsMessageExchange exchange;
            if (timeoutMS == 0) {
                exchange = this.incomingExchangesQueue.take();
            } else {
                exchange = this.incomingExchangesQueue.poll(timeoutMS, TimeUnit.MILLISECONDS);
            }
            return new MessageExchangeWrapper(exchange);
        } catch (final InterruptedException e) {
            throw new MessagingException(e);
        } finally {
            this.logger.end();
        }
    }

    @Override
    public synchronized void close() throws MessagingException {
        this.opened = false;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory() {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory.create(this);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory.createForInterface(this,
                interfaceName);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory.createForEndpoint(this,
                endpoint);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        final MessageExchangeFactory exchangeFactoryImpl = DefaultMessageExchangeFactory.createForService(this,
                serviceName);
        return exchangeFactoryImpl;
    }

    @Override
    public void send(final MessageExchange messageExchange) throws MessagingException {
        this.logger.start();

        this.realSend(((MessageExchangeWrapper) messageExchange).removeOwnership());

        this.logger.end();
    }

    private void realSend(final PetalsMessageExchange mex) throws MessagingException {

        if (mex.getTermination() != null) {
            throw new MessagingException("Exchange is terminated, reason: " + mex.getTermination());
        }
        
        if (!mex.isActive()) {
            mex.setTermination("Termination with status " + mex.getStatus());
        }

        if (mex.isNewExchange()) {
            setElectedServiceEndpoint(mex);
        } else {
            assert mex.getElectedProviderEndpoint() != null;
        }

        try {
            if (mex.isProvider()) {
                mex.setRole(Role.CONSUMER);
                this.exchangesToClientQueue.put(mex);
            } else {
                mex.setRole(Role.PROVIDER);
                this.exchangesToServerQueue.put(mex);
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    private void setElectedServiceEndpoint(final PetalsMessageExchange mex) throws MessagingException {

        final ServiceEndpoint endpoint = mex.getEndpoint();

        final PetalsServiceEndpoint ep;
        if (endpoint == null) {
            ep = this.resolveEndpoint(mex.getService(), mex.getInterfaceName());
        } else if (endpoint instanceof PetalsServiceEndpoint) {
            if (((PetalsServiceEndpoint) endpoint).getType() == EndpointType.INTERNAL) {
                ep = (PetalsServiceEndpoint) endpoint;
            } else {
                throw new MessagingException("Can't resolve external endpoints");
            }
        } else {
            ep = this.directory.resolveEndpoint(endpoint.getServiceName(), endpoint.getEndpointName());
            if (ep == null) {
                throw new MessagingException("Endpoint not available for service " + endpoint.getServiceName()
                        + " and endpoint name " + endpoint.getEndpointName());
            }
        }

        mex.setElectedProviderEndpoint(ep);
    }

    private PetalsServiceEndpoint resolveEndpoint(final QName service, final QName interfaceName)
            throws MessagingException {
        if (service != null) {
            for (final PetalsServiceEndpoint endpoint : this.directory.resolveEndpointsForService(service)) {
                if (interfaceName != null) {
                    if (ServiceEndpointHelper.matches(endpoint, interfaceName)) {
                        // let's return the first one that matches the interface
                        return endpoint;
                    }
                } else {
                    // let's return the first one
                    return endpoint;
                }
            }
            // no endpoint found
            throw new MessagingException("Endpoint not available for service " + service
                    + (interfaceName == null ? "" : (" and interface " + interfaceName)));
        }

        // note: it's ok if it's null, we will get all endpoints!
        final Collection<MockServiceEndpoint> endpoints = this.directory.resolveEndpoints(interfaceName);
        if (!endpoints.isEmpty()) {
            // let's return the first one
            return endpoints.iterator().next();
        } else {
            throw new MessagingException("Endpoint not available for "
                    + (interfaceName == null ? "any interface" : ("interface " + interfaceName)));

        }

    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange) throws MessagingException {
        this.logger.start();
        try {
            return this.sendSync(msgExchange, 0);
        } finally {
            this.logger.end();
        }
    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange, final long timeoutMS) throws MessagingException {
        this.logger.start();

        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) msgExchange;
        final PetalsMessageExchange exchange = wrapper.removeOwnership();

        // we don't need two properties like in DeliveryChannelImpl because there is only one actor that can do send
        // sync in the CDK Junit.
        exchange.setProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC, true);

        synchronized (exchange) {
            this.realSend(exchange);

            try {
                // TODO normally we should handle spurious wake-up...
                if (timeoutMS > 0) {
                    // we are sure that it's always the same object that is used in CDK Junit.
                    exchange.wait(timeoutMS);
                    // we rely on this property to know if it came back
                    if (exchange.getProperty(PetalsDeliveryChannel.PROPERTY_SENDSYNC) != null) {
                        // we timed out
                        exchange.setStatus(ExchangeStatus.ERROR);
                        exchange.setTermination("Timed out");
                        return false;
                    }
                } else {
                    exchange.wait();
                }
                wrapper.returnOwnership(exchange);
                return true;
            } catch (final InterruptedException e) {
                exchange.setStatus(ExchangeStatus.ERROR);
                exchange.setTermination("Interruption");
                throw new MessagingException(e);
            } finally {
                this.logger.end();
            }
        }
    }
}
