/**
 * Copyright (c) 2017-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.mbean;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXServiceURL;

import org.junit.rules.ExternalResource;

/**
 * The EmbeddedJmxClient Rule allows creation of an embedded Petals Component JMX client that are guaranteed to be
 * deleted when the test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public static class ... {
 *  &#064;Rule
 *  public EmbeddedJmxClient jmxClient = new EmbeddedJmxClient(jmxUrl, credentials);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      jmxClient.getTopologyServiceClient(...);
 *      ...
 *  }
 * }
 * </pre>
 */
public class EmbeddedJmxClient extends ExternalResource {

    private final String jmxUrl;

    private final String jmxUsername;

    private final String jmxPassword;

    private MBeanServerConnection jmxClient;

    private JMXConnector jmxConnector;

    private ObjectName cfgInstallerMBeanName;

    /**
     * Creates an EmbeddedJmxClient
     * 
     * @param jmxServer
     *            The JMX server to connect
     * @param username
     *            Credential username required by the JMX server
     * @param password
     *            Credential password required by the JMX server
     */
    public EmbeddedJmxClient(final EmbeddedJmxServerConnector jmxServer, final String username, final String password)
            throws MalformedObjectNameException {
        this.jmxUrl = jmxServer.getJmxUrl();
        this.jmxUsername = username;
        this.jmxPassword = password;
        this.cfgInstallerMBeanName = new ObjectName(
                "org.ow2.petals:type=custom,name=bootstrap_" + EmbeddedJmxServerConnector.COMPONENT_ID);
    }

    /**
     * Creates an EmbeddedJmxClient with default credentials:
     * <ul>
     * <li>username: {@value EmbeddedJmxServerConnector#JMX_USER}</li>
     * <li>password: {@value EmbeddedJmxServerConnector#JMX_PASSWORD}</li>
     * </ul>
     * 
     * @param jmxServer
     *            The JMX server to connect
     */
    public EmbeddedJmxClient(final EmbeddedJmxServerConnector jmxServer) throws MalformedObjectNameException {
        this(jmxServer, EmbeddedJmxServerConnector.JMX_USER, EmbeddedJmxServerConnector.JMX_PASSWORD);
    }

    @Override
    protected void before() throws Throwable {
        this.createAndConnectJmxClient();
    }

    @Override
    protected void after() {
        try {
            this.disconnectAndDestroyJmxClient();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Create and start a {@link JMXConnectorServer}
     */
    private void createAndConnectJmxClient() throws Exception {

        final JMXServiceURL url = new JMXServiceURL(this.jmxUrl);
        final String[] credentials = new String[] { this.jmxUsername, this.jmxPassword };
        final Map<String, Object> args = new HashMap<String, Object>();
        args.put(JMXConnector.CREDENTIALS, credentials);
        this.jmxConnector = JMXConnectorFactory.connect(url, args);
        this.jmxClient = this.jmxConnector.getMBeanServerConnection();
    }

    /**
     * Stop and destroy the {@link JMXConnectorServer}
     */
    private void disconnectAndDestroyJmxClient() throws Exception {
        this.jmxConnector.close();
    }

    /**
     * Set value of a component bootstrap MBean attribute
     * 
     * @param attrName
     *            The attribute name to set
     * @param value
     *            The value to set
     */
    public void setBootstrapAttribute(final String attrName, final Object value)
            throws InstanceNotFoundException, AttributeNotFoundException, InvalidAttributeValueException,
            MBeanException, ReflectionException, IOException {
        this.jmxClient.setAttribute(this.cfgInstallerMBeanName, new Attribute(attrName, value));
    }

    /**
     * Get value as {@link String} of a component bootstrap MBean attribute
     * 
     * @param attrName
     *            The attribute name to get
     * @return The value
     */
    public String getBootstrapAttributeAsString(final String attrName) throws AttributeNotFoundException,
            InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        return (String) this.jmxClient.getAttribute(this.cfgInstallerMBeanName, attrName);
    }

    /**
     * Get value as {@link int} of a component bootstrap MBean attribute
     * 
     * @param attrName
     *            The attribute name to get
     * @return The value
     */
    public int getBootstrapAttributeAsInt(final String attrName) throws AttributeNotFoundException,
            InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        return (Integer) this.jmxClient.getAttribute(this.cfgInstallerMBeanName, attrName);
    }

    /**
     * Get value as {@link long} of a component bootstrap MBean attribute
     * 
     * @param attrName
     *            The attribute name to get
     * @return The value
     */
    public long getBootstrapAttributeAsLong(final String attrName) throws AttributeNotFoundException,
            InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        return (Long) this.jmxClient.getAttribute(this.cfgInstallerMBeanName, attrName);
    }

    /**
     * Get value as {@link boolean} of a component bootstrap MBean attribute
     * 
     * @param attrName
     *            The attribute name to get
     * @return The value
     */
    public boolean getBootstrapAttributeAsBoolean(final String attrName) throws AttributeNotFoundException,
            InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        return (Boolean) this.jmxClient.getAttribute(this.cfgInstallerMBeanName, attrName);
    }
}
