/**
 * Copyright (c) 2014-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.rule;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.TimeoutException;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.apache.commons.io.FileUtils;
import org.junit.Rule;
import org.junit.rules.ExternalResource;
import org.junit.rules.TemporaryFolder;
import org.ow2.easywsdl.wsdl.WSDLFactory;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.WSDLReader;
import org.ow2.easywsdl.wsdl.api.WSDLWriter;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.junit.Assert;
import org.ow2.petals.component.framework.junit.Component;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.ResponseMessage;
import org.ow2.petals.component.framework.junit.StatusMessage;
import org.ow2.petals.component.framework.junit.TestMessageExchangeFactory;
import org.ow2.petals.component.framework.junit.impl.ComponentConfiguration;
import org.ow2.petals.component.framework.junit.impl.ComponentInstanceImpl;
import org.ow2.petals.component.framework.junit.impl.ServiceConfiguration;
import org.ow2.petals.component.framework.junit.impl.message.AbstractWrappedToComponentMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedFaultFromProviderMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedRequestFromConsumerMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedResponseFromProviderMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedStatusFromConsumerMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedStatusFromProviderMessage;
import org.ow2.petals.component.framework.junit.impl.mock.MockEndpointDirectory;
import org.ow2.petals.component.framework.junit.impl.mock.MockServiceEndpoint;
import org.ow2.petals.component.framework.junit.impl.mock.TestMessageExchangeFactoryImpl;
import org.ow2.petals.component.framework.junit.mbean.EmbeddedJmxServerConnector;
import org.ow2.petals.component.framework.logger.StepLogHelper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.log.handler.PetalsFileHandler;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * The {@link ComponentUnderTest} {@link Rule} allows creation of component instance of the current component project.
 * It is guaranteed to be deleted when the test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public class MyTestSuite {
 *  &#064;Rule
 *  public ComponentUnderTest componentUnderTest = new ComponentUnderTest(...);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      componentUnderTest.start();
 *      componentUnderTest.installService(...);
 *      ...
 *  }
 * }
 * </pre>
 * 
 * If a file logging.properties is present in the classpath, it is used for the logs.
 */
public class ComponentUnderTest extends ExternalResource implements Component {

    static {

        // let's initialize logging stuffs, it has to be done here to happen before the rest.
        Level.initialize();

        final InputStream logConfig = ComponentUnderTest.class.getResourceAsStream("/logging.properties");
        if (logConfig != null) {
            try {
                LogManager.getLogManager().readConfiguration(logConfig);
            } catch (final SecurityException | IOException e) {
                fail(e.getMessage());
            }
        }
    }

    /**
     * A temporary folder that will be used to define the root of the file-system used by the component instance.
     */
    private final TemporaryFolder temporaryFolder = new TemporaryFolder();

    /**
     * Set of the component JBI descriptor parameters that must be generated on test case set-up (see
     * {@link ComponentUnderTest#before()})
     */
    private final Map<QName, ParameterGenerator> parameterGenerators = new HashMap<>();

    /**
     * The component instance configuration
     */
    private final ComponentConfiguration componentConfiguration;

    /**
     * The component instance
     */
    private ComponentInstanceImpl componentInst = null;

    /**
     * Flag controlling the component must be installed on setup
     */
    private final boolean componentMustBeInstalled;

    /**
     * Flag controlling if the component must be started on setup
     */
    private final boolean componentMustBeStarted;

    /**
     * The directory where the component is installed
     */
    private File componentInstallationDir = null;

    private TestMessageExchangeFactoryImpl factory;

    /**
     * Service configurations factory creating the service configuration to deploy when the component will be
     * installed/started
     */
    private final Map<String, ServiceConfigurationFactory> serviceCfgFactoryRegisteredForDeploiement = new HashMap<>();

    /**
     * Native service configurations factory creating the service configuration to deploy when the component will be
     * installed/started
     */
    private final Map<String, NativeServiceConfigurationFactory> nativeServiceCfgFactoryRegisteredForDeploiement = new HashMap<>();

    /**
     * Service configurations deployed
     */
    private final Map<String, ServiceConfiguration> serviceConfigurations = new HashMap<>();

    /**
     * external endponit to register
     */
    private final Set<MockServiceEndpoint> externalServiceProviderToActivate = new HashSet<>();

    /**
     * The files to add in the component file hierarchy (hack needed by Petals BC SOAP to load the addressing module)
     */
    private final Map<String, String> filesToAdd = new HashMap<>();

    private final Set<Handler> logHandlers = new HashSet<>();

    private Logger componentLogger;

    private Logger externalServiceLogger;

    private MockEndpointDirectory directory;

    private EmbeddedJmxServerConnector embeddedJmxSrv = null;

    /**
     * 
     * Create a component under test. It is automatically installed and started
     * 
     */
    public ComponentUnderTest() {
        this(new ComponentConfiguration("UnderTestWithDefaultConf"), true, true);
    }

    /**
     * Create a component under test controllong its inastallation and start-up states
     * 
     * @param installesComponent
     *            Flag controlling the component must be installed on setup
     * @param startsComponent
     *            Flag controlling if the component must be started on setup
     */
    public ComponentUnderTest(final boolean installesComponent, final boolean startsComponent) {
        this(new ComponentConfiguration("UnderTestWithDefaultConf"), installesComponent, startsComponent);
    }

    public ComponentUnderTest(final ComponentConfiguration configuration) {
        this(configuration, true, true);
    }

    /**
     * 
     * @param componentConfiguration
     *            The component configuration
     * @param installesComponent
     *            Flag controlling the component must be installed on setup
     * @param startsComponent
     *            Flag controlling if the component must be started on setup
     */
    public ComponentUnderTest(final ComponentConfiguration componentConfiguration, final boolean installesComponent,
            final boolean startsComponent) {
        this.componentConfiguration = componentConfiguration;
        assertTrue("Unable to start a component not initialized", (startsComponent && installesComponent)
                || !startsComponent);
        this.componentMustBeInstalled = installesComponent;
        this.componentMustBeStarted = startsComponent;
    }

    @Override
    protected void before() throws Throwable {
        this.create();
    }

    public void create() throws Throwable {
        this.temporaryFolder.create();
        
        this.componentInstallationDir = this.temporaryFolder.newFolder();

        this.generateParameters();

        System.setProperty(PetalsFileHandler.PETALS_LOG_DIR_PROPERTY_NAME, new File(this.componentInstallationDir,
                "log").getAbsolutePath());

        for (final Entry<String, String> fileToAddEntry : this.filesToAdd.entrySet()) {
            final File srcFile = new File(fileToAddEntry.getKey());
            final File dstFile = new File(new File(this.componentInstallationDir, fileToAddEntry.getValue()),
                    srcFile.getName());
            FileUtils.copyFile(srcFile, dstFile);
        }

        this.directory = new MockEndpointDirectory();

        for (final MockServiceEndpoint endpoint : this.externalServiceProviderToActivate) {
            this.directory.activateEndpoint(endpoint);
        }

        this.externalServiceLogger = Logger.getLogger("org.ow2.petals.component.framework.external-services");
        this.componentLogger = Logger
                .getLogger("org.ow2.petals.component.framework." + componentConfiguration.getConfigurationName());

        for (final Handler handler : logHandlers) {
            this.externalServiceLogger.addHandler(handler);
            this.componentLogger.addHandler(handler);
            // TODO support other logger created by the component?!
        }

        this.componentInst = new ComponentInstanceImpl(this.directory, this.componentConfiguration,
                this.componentInstallationDir, this.componentLogger);
        this.componentInst.setEmbeddedJmxSrv(this.embeddedJmxSrv);
        this.componentInst.load();

        if (this.componentMustBeInstalled) {
            this.install();

            this.deployServiceConfigurations();

            if (this.componentMustBeStarted) {
                this.start();
            }

            this.factory = new TestMessageExchangeFactoryImpl(this.directory, this.componentInst.getLogger());
        }
    }

    @Override
    protected void after() {
        this.delete();
    }

    public void delete() {
        if (this.isStarted()) {
            this.stop();
        }

        if (this.isInstalled()) {
            this.undeployAllServices();
            this.shutdown();
        }

        System.clearProperty(PetalsFileHandler.PETALS_LOG_DIR_PROPERTY_NAME);

        this.temporaryFolder.delete();

        // because loggers are long-lived in the JVM, we have to remove that if not handlers will be readded in the next
        // run of a test
        for (final Handler handler : logHandlers) {
            this.externalServiceLogger.removeHandler(handler);
            this.componentLogger.removeHandler(handler);
            // TODO support other logger created by the component?!
        }
    }

    @Override
    public ComponentUnderTest addLogHandler(final Handler logHandler) {
        this.logHandlers.add(logHandler);
        return this;
    }

    @Override
    public ComponentUnderTest addEmbeddedJmxSrv(final EmbeddedJmxServerConnector embeddedJmxSrvCon) {
        this.embeddedJmxSrv = embeddedJmxSrvCon;
        return this;
    }

    @Override
    public String getComponentName() {
        return this.componentInst.getComponentName();
    }

    @Override
    public AbstractComponent getComponentObject() {
        return this.componentInst.getComponentObject();
    }

    @Override
    public File getBaseDirectory() {
        return this.componentInstallationDir;
    }

    @Override
    public void install() {
        this.componentInst.install();
    }

    @Override
    public void deployServiceConfigurations() throws DeploymentException {

        for (final Entry<String, ServiceConfigurationFactory> serviceToDeploy : this.serviceCfgFactoryRegisteredForDeploiement
                .entrySet()) {
            this.deployService(serviceToDeploy.getKey(), serviceToDeploy.getValue());
        }

        for (final Entry<String, NativeServiceConfigurationFactory> serviceToDeploy : this.nativeServiceCfgFactoryRegisteredForDeploiement
                .entrySet()) {
            this.registerNativeService(serviceToDeploy.getKey(), serviceToDeploy.getValue());
        }
    }

    @Override
    public boolean isInstalled() {
        return this.componentInst != null && this.componentInst.isInstalled();
    }

    @Override
    public void start() {
        this.componentInst.start();
    }

    @Override
    public boolean isStarted() {
        return this.componentInst != null && this.componentInst.isStarted();
    }

    @Override
    public void stop() {
        this.componentInst.stop();
    }

    @Override
    public void shutdown() {
        this.componentInst.shutdown();
    }

    @Override
    public ComponentUnderTest registerServiceToDeploy(final String serviceCfgId,
            final ServiceConfigurationFactory serviceConfigurationFactory) {
        Assert.assertFalse("Service configuration already registered",
                this.serviceCfgFactoryRegisteredForDeploiement.containsKey(serviceCfgId));
        this.serviceCfgFactoryRegisteredForDeploiement.put(serviceCfgId, serviceConfigurationFactory);
        return this;
    }

    @Override
    public void deployService(final String serviceCfgId, final ServiceConfigurationFactory serviceConfigurationFactory)
            throws DeploymentException {
        ServiceConfiguration serviceConfiguration;
        try {
            serviceConfiguration = serviceConfigurationFactory.create();
        } catch (final Exception e) {
            throw new DeploymentException(e);
        }
        this.deployService(serviceCfgId, serviceConfiguration);
    }

    @Override
    public void deployService(final String suName, final ServiceConfiguration serviceConfiguration)
            throws DeploymentException {
        this.serviceConfigurations.put(suName, serviceConfiguration);
        this.componentInst.deployService(suName, serviceConfiguration);
    }

    @Override
    public void initService(final String suName) throws DeploymentException {
        this.componentInst.initService(suName);
    }

    @Override
    public void startService(final String suName) throws DeploymentException {
        this.componentInst.startService(suName);
    }

    @Override
    public void stopService(final String suName) throws DeploymentException {
        this.componentInst.stopService(suName);
    }

    @Override
    public void shutdownService(final String suName) throws DeploymentException {
        this.componentInst.shutdownService(suName);
    }

    @Override
    public boolean isServiceDeployed(final String suName) {
        return this.componentInst.isServiceDeployed(suName);
    }

    @Override
    public void undeployService(final String serviceConfigurationId) {
        final ServiceConfiguration serviceConfiguration = this.serviceConfigurations.remove(serviceConfigurationId);
        if (serviceConfiguration != null) {
            this.componentInst.undeployService(serviceConfigurationId);
        }
    }

    @Override
    public void undeployAllServices() {
        this.componentInst.undeployAllServices();
        this.serviceConfigurations.clear();
    }

    @Override
    public ComponentUnderTest registerNativeServiceToDeploy(final String serviceCfgId,
            final NativeServiceConfigurationFactory serviceConfigurationFactory) {
        Assert.assertFalse("Service configuration already registered",
                this.nativeServiceCfgFactoryRegisteredForDeploiement.containsKey(serviceCfgId));
        this.nativeServiceCfgFactoryRegisteredForDeploiement.put(serviceCfgId, serviceConfigurationFactory);
        return this;
    }

    @Override
    public void registerNativeService(final String serviceCfgId,
            final NativeServiceConfigurationFactory serviceConfigurationFactory) {
        Assert.assertFalse("Service configuration already registered",
                this.serviceConfigurations.containsKey(serviceCfgId));
        final ServiceConfiguration serviceConfiguration = serviceConfigurationFactory.create(this.componentInst
                .getNativeEndpointName(serviceConfigurationFactory.getNativeService()));
        this.serviceConfigurations.put(serviceCfgId, serviceConfiguration);
    }

    @Override
    public ComponentUnderTest registerExternalServiceProvider(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        return this.registerExternalServiceProvider(endpointName, serviceName, interfaceName, null);
    }

    @Override
    public ComponentUnderTest registerExternalServiceProvider(final String endpointName, final QName serviceName,
            final QName interfaceName, final URL wsdlUrl) {

        if (wsdlUrl != null) {
            try {
                final WSDLReader wsdlReader = WSDLFactory.newInstance().newWSDLReader();
                final Description wsdlDescr = wsdlReader.read(wsdlUrl);
                final WSDLWriter wsdlWriter = WSDLFactory.newInstance().newWSDLWriter();

                this.externalServiceProviderToActivate.add(new MockServiceEndpoint(endpointName, serviceName,
                        interfaceName, wsdlWriter.getDocument(wsdlDescr)));
            } catch (final URISyntaxException | IOException | WSDLException e) {
                throw new UncheckedException(e);
            }
        } else {
            this.externalServiceProviderToActivate
                    .add(new MockServiceEndpoint(endpointName, serviceName, interfaceName));
        }
        return this;
    }

    @Override
    public ComponentUnderTest activateExternalServiceProvider(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        if (this.directory != null) {
            this.directory.activateEndpoint(new MockServiceEndpoint(endpointName, serviceName, interfaceName));
        } else {
            Assert.fail("Can't activate endpoint before the rule has been initialised, maybe use register instead?");
        }
        return this;
    }

    private void pushToProvider(final Message message, final boolean sync) {
        assert message instanceof AbstractWrappedToComponentMessage;
        assert message instanceof RequestMessage || message instanceof StatusMessage;

        final PetalsMessageExchange msgExchange = (PetalsMessageExchange) message.getMessageExchange();

        // it's the first step of an exchange between a test and the component
        // let's setup the flow attributes
        if (msgExchange.isNewExchange()) {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();
            if (flowAttributes == null) {
                flowAttributes = PetalsExecutionContext.initFlowAttributes();
            }
            FlowAttributesExchangeHelper.setFlowAttributes(msgExchange, flowAttributes);
        }

        this.componentInst.pushMessageInDeliveryChannel(msgExchange, sync);
    }

    @Override
    public void pushRequestToProvider(final RequestMessage message) {
        pushToProvider(message, false);
    }

    @Override
    public void pushRequestToProvider(RequestMessage message, boolean sync) {
        pushToProvider(message, sync);
    }

    @Override
    public void pushStatusToProvider(StatusMessage message) {
        pushToProvider(message, false);
    }

    @Override
    public void pushStatusToProvider(StatusMessage message, boolean sync) {
        pushToProvider(message, sync);
    }

    @Override
    public ResponseMessage pollResponseFromProvider() throws TimeoutException {
        return pollResponseFromProvider(DEFAULT_TIMEOUT);
    }

    @Override
    public ResponseMessage pollResponseFromProvider(long timeout) throws TimeoutException {
        final PetalsMessageExchange msgExchange = pollFromProvider(timeout);

        if (msgExchange.getFault() != null) {
            return new WrappedFaultFromProviderMessage(msgExchange);
        } else if (msgExchange.getMessage(Exchange.OUT_MESSAGE_NAME) != null) {
            return new WrappedResponseFromProviderMessage(msgExchange);
        } else {
            throw new UncheckedException("A response message must have either an out or a fault.",
                    msgExchange.getError());
        }
    }

    @Override
    public StatusMessage pollStatusFromProvider() throws TimeoutException {
        return pollStatusFromProvider(DEFAULT_TIMEOUT);
    }

    @Override
    public StatusMessage pollStatusFromProvider(long timeout) throws TimeoutException {
        final PetalsMessageExchange msgExchange = pollFromProvider(timeout);

        return new WrappedStatusFromProviderMessage(msgExchange);
    }

    private PetalsMessageExchange pollFromProvider(final long timeout) throws TimeoutException {
        final PetalsMessageExchange msgExchange = this.componentInst.pollFromProvider(timeout);

        // the poller (a unit test or a service implementation) will have the correct exec context
        final FlowAttributes fa = FlowAttributesExchangeHelper.getFlowAttributes(msgExchange);
        PetalsExecutionContext.putFlowAttributes(fa);

        return msgExchange;
    }

    @Override
    public void clearResponsesFromProvider() {
        this.componentInst.clearResponsesFromProvider();
    }

    @Override
    public int getResponsesFromProviderCount() {
        return this.componentInst.getResponsesFromProviderCount();
    }

    @Override
    public RequestMessage pollRequestFromConsumer() throws TimeoutException {
        return pollRequestFromConsumer(DEFAULT_TIMEOUT);
    }

    private PetalsMessageExchange pollFromConsumerAndLog(final long timeout) throws TimeoutException {
        final PetalsMessageExchange msg = this.componentInst.pollFromConsumer(timeout);

        // this is a new step for the external provider!
        if (msg != null) {

            final FlowAttributes fa = FlowAttributesExchangeHelper.getFlowAttributes(msg);
            PetalsExecutionContext.putFlowAttributes(fa);

            if (msg.isNewExchange()) {
                final FlowAttributes current = PetalsExecutionContext.nextFlowStepId();
                StepLogHelper.addMonitStartTrace(externalServiceLogger, msg, current, fa);
            }
        }

        return msg;
    }

    @Override
    public RequestMessage pollRequestFromConsumer(long timeout) throws TimeoutException {
        final PetalsMessageExchange msg = pollFromConsumerAndLog(timeout);
        return msg != null ? new WrappedRequestFromConsumerMessage(msg) : null;
    }

    @Override
    public StatusMessage pollStatusFromConsumer() throws TimeoutException {
        return pollStatusFromConsumer(DEFAULT_TIMEOUT);
    }

    @Override
    public StatusMessage pollStatusFromConsumer(long timeout) throws TimeoutException {
        final PetalsMessageExchange msg = pollFromConsumerAndLog(timeout);
        return msg != null ? new WrappedStatusFromConsumerMessage(msg) : null;
    }

    @Override
    public void clearRequestsFromConsumer() {
        this.componentInst.clearRequestsFromConsumer();
    }

    @Override
    public int getRequestsFromConsumerCount() {
        return this.componentInst.getRequestsFromConsumerCount();
    }

    @Override
    public int getExchangesInDeliveryChannelCount() {
        return this.componentInst.getExchangesInDeliveryChannelCount();
    }

    private void pushToConsumerAndLog(final Message message, final boolean sync) {
        assert message instanceof AbstractWrappedToComponentMessage;
        assert message instanceof ResponseMessage || message instanceof StatusMessage;

        final FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();
        StepLogHelper.addMonitEndOrFailureTrace(externalServiceLogger, message.getMessageExchange(), flowAttributes);

        this.componentInst.pushMessageInDeliveryChannel((PetalsMessageExchange) message.getMessageExchange(), sync);
    }

    @Override
    public void pushResponseToConsumer(final ResponseMessage message) {
        pushToConsumerAndLog(message, false);
    }

    @Override
    public void pushResponseToConsumer(final ResponseMessage message, final boolean sync) {
        pushToConsumerAndLog(message, sync);
    }

    @Override
    public void pushStatusToConsumer(final StatusMessage message) {
        pushToConsumerAndLog(message, false);
    }

    @Override
    public void pushStatusToConsumer(final StatusMessage message, final boolean sync) {
        pushToConsumerAndLog(message, sync);
    }

    @Override
    public ServiceConfiguration getServiceConfiguration(final String suName) {
        return this.serviceConfigurations.get(suName);
    }

    @Override
    public String getNativeEndpointName(final QName service) {
        return this.componentInst.getNativeEndpointName(service);
    }

    @Override
    public File getServiceConfigurationInstallDir(final String suName) {
        return this.componentInst.getServiceConfigurationInstallDir(suName);
    }

    @Override
    public ComponentUnderTest setParameter(final QName paramName, final String value) {
        this.componentConfiguration.setParameter(paramName, value);
        return this;
    }

    @Override
    public ComponentUnderTest setParameter(final QName paramName, final ParameterGenerator paramGenerator) {
        this.parameterGenerators.put(paramName, paramGenerator);
        return this;
    }

    private void generateParameters() throws Exception {
        for (final Entry<QName, ParameterGenerator> parameter : this.parameterGenerators.entrySet()) {
            this.setParameter(parameter.getKey(), parameter.getValue().generate());
        }
    }

    @Override
    public ComponentUnderTest addFile(final String fileToAdd, final String destInComponentDir) {
        this.filesToAdd.put(fileToAdd, destInComponentDir);
        return this;
    }

    @Override
    public TestMessageExchangeFactory getTestMessageExchangeFactory() {
        return this.factory;
    }

    @Override
    public MockEndpointDirectory getEndpointDirectory() {
        return this.directory;
    }
}
