/**
 * Copyright (c) 2015-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.helpers;

import java.util.concurrent.TimeoutException;

import javax.jbi.messaging.ExchangeStatus;

import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.junit.Assert;
import org.ow2.petals.component.framework.junit.Component;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.ResponseMessage;
import org.ow2.petals.component.framework.junit.StatusMessage;
import org.ow2.petals.component.framework.junit.impl.message.StatusToProviderMessage;

/**
 * Simple wrapper that provides some high-level methods exploiting {@link ServiceProviderImplementation} and
 * {@link MessageChecks} to send and receives messages to/from a {@link Component}.
 * 
 * @author Victor Noel
 *
 */
public class SimpleComponent {

    public static final long DEFAULT_SEND_AND_RECEIVE_TIMEOUT = Component.DEFAULT_TIMEOUT;

    private final Component component;

    public SimpleComponent(final Component component) {
        this.component = component;
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link StatusMessage} from the
     * component under test.
     * 
     */
    public StatusMessage sendAndGetStatus(final RequestMessage request, final ServiceProviderImplementation service)
            throws Exception {
        return sendAndGetStatus(request, service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Send a {@link RequestMessage} to the component under test and wait for the {@link StatusMessage} from the
     * component under test.
     * 
     */
    public StatusMessage sendAndGetStatus(final RequestMessage request) throws TimeoutException {
        return sendAndGetStatus(request, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link StatusMessage} from the
     * component under test.
     * 
     */
    public StatusMessage sendAndGetStatus(final RequestMessage request, final ServiceProviderImplementation service,
            final long timeoutForth, final long timeoutBack) throws Exception {

        PetalsExecutionContext.clear();

        component.pushRequestToProvider(request);

        receiveRequestAsExternalProvider(service, timeoutForth, false);

        return component.pollStatusFromProvider(timeoutBack);
    }

    /**
     * Send a {@link RequestMessage} to the component under test and wait for the {@link StatusMessage} from the
     * component under test.
     * 
     */
    public StatusMessage sendAndGetStatus(final RequestMessage request, final long timeoutBack)
            throws TimeoutException {

        PetalsExecutionContext.clear();

        component.pushRequestToProvider(request);

        return component.pollStatusFromProvider(timeoutBack);
    }

    /**
     * Send a {@link RequestMessage} to the component under test and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * One must use {@link #sendStatus(StatusMessage, boolean)} after using this one!!!
     */
    public ResponseMessage sendAndGetResponse(final RequestMessage request) throws TimeoutException {
        return sendAndGetResponse(request, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * <p>Send à {@link RequestMessage} to the component under test and wait for the {@link ResponseMessage} from the
     * component under test.</p>
     * 
     * <p>One must use {@link #sendStatus(StatusMessage, boolean)} after using this one!!!</p>
     * 
     */
    public ResponseMessage sendAndGetResponse(final RequestMessage request, final long timeoutBack)
            throws TimeoutException {

        PetalsExecutionContext.clear();

        component.pushRequestToProvider(request);

        return component.pollResponseFromProvider(timeoutBack);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * One must use {@link #sendStatus(StatusMessage, ServiceProviderImplementation, long)} after using this one!!!
     */
    public ResponseMessage sendAndGetResponse(final RequestMessage request, final ServiceProviderImplementation service)
            throws Exception {
        return sendAndGetResponse(request, service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * One must use {@link #sendStatus(StatusMessage, ServiceProviderImplementation, long)} after using this one!!!
     * 
     */
    public ResponseMessage sendAndGetResponse(final RequestMessage request, final ServiceProviderImplementation service,
            final long timeoutForth, final long timeoutBack) throws Exception {

        PetalsExecutionContext.clear();

        component.pushRequestToProvider(request);

        receiveRequestAsExternalProvider(service, timeoutForth, false);

        return component.pollResponseFromProvider(timeoutBack);
    }

    /**
     * <p>
     * Send a {@link RequestMessage}, as an external service provider, to the component under test and wait for the
     * {@link StatusMessage} from the component under test.
     * </p>
     * <p>
     * To use in the service implementation of {@link #receiveAsExternalProvider(ServiceProviderImplementation)}.
     * </p>
     * 
     */
    public StatusMessage sendAndGetStatusAsExternalProvider(final RequestMessage request) throws TimeoutException {
        return sendAndGetStatusAsExternalProvider(request, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * <p>
     * Send a {@link RequestMessage}, as an external service provider, to the component under test and wait for the
     * {@link StatusMessage} from the component under test.
     * </p>
     * <p>
     * To use in the service implementation of {@link #receiveAsExternalProvider(ServiceProviderImplementation)}.
     * </p>
     * 
     */
    public StatusMessage sendAndGetStatusAsExternalProvider(final RequestMessage request, final long timeoutBack)
            throws TimeoutException {

        component.pushRequestToProvider(request);

        return component.pollStatusFromProvider(timeoutBack);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * Then check the {@link ResponseMessage} with the provided {@link MessageChecks} and then send back the provided
     * {@link ExchangeStatus}.
     * 
     */
    public void sendAndCheckResponseAndSendStatus(final RequestMessage request,
            final ServiceProviderImplementation service, final MessageChecks checks, final ExchangeStatus status)
            throws Exception {
        sendAndCheckResponseAndSendStatus(request, service, checks, status, DEFAULT_SEND_AND_RECEIVE_TIMEOUT,
                DEFAULT_SEND_AND_RECEIVE_TIMEOUT, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * Then check the {@link ResponseMessage} with the provided {@link MessageChecks} and then send back the provided
     * {@link Exception} as an error.
     * 
     */
    public void sendAndCheckResponseAndSendStatus(final RequestMessage request,
            final ServiceProviderImplementation service, final MessageChecks checks, final Exception error)
            throws Exception {
        sendAndCheckResponseAndSendStatus(request, service, checks, error, DEFAULT_SEND_AND_RECEIVE_TIMEOUT,
                DEFAULT_SEND_AND_RECEIVE_TIMEOUT, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * Then check the {@link ResponseMessage} with the provided {@link MessageChecks} and then send back the provided
     * {@link ExchangeStatus}.
     * 
     */
    public void sendAndCheckResponseAndSendStatus(final RequestMessage request,
            final ServiceProviderImplementation service, final MessageChecks checks, final ExchangeStatus status,
            final long timeoutRequest, final long timeoutResponse, final long timeoutStatus) throws Exception {
        final ResponseMessage response = sendAndGetResponse(request, service, timeoutRequest, timeoutResponse);
        checks.checks(response);
        sendStatus(new StatusToProviderMessage(response, status), service, timeoutStatus);
    }

    /**
     * Send a {@link RequestMessage} to the component under test, answer the component under test request to an external
     * service with the provided {@link ServiceProviderImplementation} and wait for the {@link ResponseMessage} from the
     * component under test.
     * 
     * Then check the {@link ResponseMessage} with the provided {@link MessageChecks} and then send back the provided
     * {@link Exception} as an error.
     * 
     */
    public void sendAndCheckResponseAndSendStatus(final RequestMessage request,
            final ServiceProviderImplementation service, final MessageChecks checks, final Exception error,
            final long timeoutRequest, final long timeoutResponse, final long timeoutStatus) throws Exception {
        final ResponseMessage response = sendAndGetResponse(request, service, timeoutRequest, timeoutResponse);
        checks.checks(response);
        sendStatus(new StatusToProviderMessage(response, error), service, timeoutStatus);
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, long)} to send the closing DONE status.
     */
    public void sendDoneStatus(final ResponseMessage response) {
        sendStatus(new StatusToProviderMessage(response, ExchangeStatus.DONE), true);
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, long)} to send the closing status.
     */
    public void sendStatus(final StatusMessage status, final boolean wait) {

        PetalsExecutionContext.clear();

        component.pushStatusToProvider(status, wait);
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, ServiceProviderImplementation, long, long)} to send
     * the closing status.
     */
    public void sendStatus(final StatusMessage status, final ServiceProviderImplementation service) throws Exception {
        sendStatus(status, service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, ServiceProviderImplementation, long, long)} to send
     * the closing status.
     */
    public void sendStatus(final StatusMessage status, final ServiceProviderImplementation service,
            final long timeoutBack) throws Exception {

        PetalsExecutionContext.clear();

        component.pushStatusToProvider(status, false);

        if (service.statusExpected()) {
            receiveStatusAsExternalProvider(service, timeoutBack);
        }
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, ServiceProviderImplementation, long, long)} to send
     * the closing DONE status.
     */
    public void sendDoneStatus(final ResponseMessage response, final ServiceProviderImplementation service)
            throws Exception {
        sendDoneStatus(response, service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Use after using {@link #sendAndGetResponse(RequestMessage, ServiceProviderImplementation, long, long)} to send
     * the closing DONE status.
     */
    public void sendDoneStatus(final ResponseMessage response, final ServiceProviderImplementation service,
            final long timeoutBack) throws Exception {
        sendStatus(new StatusToProviderMessage(response, ExchangeStatus.DONE), service, timeoutBack);
    }

    /**
     * Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.
     */
    public void receiveAsExternalProvider(final ServiceProviderImplementation service) throws Exception {
        receiveAsExternalProvider(service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT, false);
    }

    /**
     * Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.
     */
    public void receiveAsExternalProvider(final ServiceProviderImplementation service, final long timeout)
            throws Exception {
        receiveAsExternalProvider(service, timeout, false);
    }

    /**
     * Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.
     * 
     * @param wait
     *            if <code>true</code> and either the external service does not expect a status or if it answers with
     *            one, then wait until the message has been processed by the component under test before returning
     */
    public void receiveAsExternalProvider(final ServiceProviderImplementation service, final boolean wait)
            throws Exception {
        receiveAsExternalProvider(service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT, wait);
    }

    /**
     * <p>Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.</p>
     * 
     * <p>It covers the whole exchange: Request -&gt; Response -&gt; Status (if expected by the external service)</p>
     * 
     * @param wait
     *            if <code>true</code> and either the external service does not expect a status or if it answers with
     *            one, then wait until the message has been processed by the component under test before returning
     */
    public void receiveAsExternalProvider(final ServiceProviderImplementation service, final long timeout,
            final boolean wait) throws Exception {
        // let's wait only if we don't have to wait for an acknowledgement after
        receiveRequestAsExternalProvider(service, timeout, !service.statusExpected() && wait);
        if (service.statusExpected()) {
            receiveStatusAsExternalProvider(service, timeout);
        }
    }

    /**
     * <p>Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.</p>
     * 
     * <p>This covers the first step of an exchange (Request -&gt; Response).</p>
     * 
     * @param wait
     *            if <code>true</code> wait until the message has been processed by the component under test
     */
    public void receiveResponseAsExternalProvider(final ServiceProviderImplementation service, final boolean wait)
            throws Exception {
        receiveRequestAsExternalProvider(service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT, wait);
    }

    /**
     * <p>Answer the component under test request to an external service with the provided
     * {@link ServiceProviderImplementation}.</p>
     * 
     * <p>This covers the first step of an exchange (Request -&gt; Response).</p>
     * 
     * @param wait
     *            if <code>true</code> wait until the message has been processed by the component under test
     */
    public void receiveRequestAsExternalProvider(final ServiceProviderImplementation service, final long timeout,
            final boolean wait) throws Exception {
        final RequestMessage requestFromConsumer = component.pollRequestFromConsumer(timeout);
        Assert.assertNotNull(requestFromConsumer);
        // TODO inOptOut is not really supported...
        final Message responseToConsumer = service.provides(requestFromConsumer);
        if (responseToConsumer instanceof ResponseMessage) {
            component.pushResponseToConsumer((ResponseMessage) responseToConsumer, wait);
        } else if (responseToConsumer instanceof StatusMessage) {
            if (service.statusExpected()) {
                Assert.fail(
                        "A response is expected but the external service implementation returned a status message");
            }
            component.pushStatusToConsumer((StatusMessage) responseToConsumer, wait);
        } else {
            Assert.fail(String.format("Unexpected message type '%s' for: %s", responseToConsumer.getClass(),
                    responseToConsumer));
        }
    }

    /**
     * Receive the component under test status to an external service with the provided
     * {@link ServiceProviderImplementation}.
     * 
     * This covers the third step of an exchange (Status).
     */
    public void receiveStatusAsExternalProvider(final ServiceProviderImplementation service) throws Exception {
        receiveStatusAsExternalProvider(service, DEFAULT_SEND_AND_RECEIVE_TIMEOUT);
    }

    /**
     * Receive the component under test status to an external service with the provided
     * {@link ServiceProviderImplementation}.
     * 
     * This covers the third step of an exchange (Status).
     */
    public void receiveStatusAsExternalProvider(final ServiceProviderImplementation service, final long timeout)
            throws Exception {
        Assert.assertTrue("The external service implementation does not expect a status message",
                service.statusExpected());
        final StatusMessage status = component.pollStatusFromConsumer(timeout);
        Assert.assertNotNull(status);
        service.handleStatus(status);
    }

}
